/*
 * buffer.c
 *
 * routines of input
 * Copyright (c) 1988, 89, 90, 91, 92, 93 Miguel Santana
 * Copyright (c) 1995, 96, 97, 98 Akim Demaille, Miguel Santana
 * $Id: buffer.h,v 1.28 1998/03/03 11:47:05 demaille Exp $
 */

/*
 * This file is part of a2ps.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * $Id: buffer.h,v 1.28 1998/03/03 11:47:05 demaille Exp $
 */

#ifndef _BUFFER_H_
#define _BUFFER_H_
#include "xobstack.h"

/****************************************************************/
/*		 	struct buffer Service routines 		*/
/****************************************************************/
/*
 * How the eol should be handled
 */
enum eol_e {
  eol_r, 	/* \r only */
  eol_n,	/* \n 	*/
  eol_rn,	/* \r\n	*/
  eol_nr,	/* \n\r */
  eol_auto	/* Any of the previous */
};

const char * eol_to_string PARAMS ((enum eol_e eol));
enum eol_e option_string_to_eol PARAMS ((const char * option,
				       const char * arg));

/* 
 * The buffer structure, which encloses a lower case version to
 * speed up case insensitive parsing
 * There are two types:
 * - stream != NULL : input comes from a stream
 * - stream == NULL : input comes from stream_string, read line by line.
 */
/* I suppose unions are better, but I'm not used to them... */
struct buffer {
  FILE * stream;
  const uchar * stream_string;
  size_t offset;	/* Used when a buffer string	*/
  uchar * content;	/* Exactly what is read	*/
  enum eol_e eol;		/* What is an end of line?		*/  
  int lower_case;
  uchar * value;	/* if LOWER_CASE, then lower case of content */
  size_t line;		/* Num of the current line */
  size_t allocsize;	/* Used to know how big lower_case is */
  size_t len;
  size_t curr;
  struct obstack * input_stack;
};

struct buffer * buffer_stream_new PARAMS ((FILE * stream,
					   enum eol_e eol));

struct buffer * buffer_string_new PARAMS ((const uchar * string,
					   enum eol_e eol));
void buffer_set_lower_case PARAMS ((struct buffer * buffer,
				    int lower_case));
void buffer_self_print PARAMS ((struct buffer * buffer, FILE * stream));

void buffer_free PARAMS ((struct buffer * buffer));

void buffer_get PARAMS ((struct buffer * buffer));

#define buffer_is_empty(Buf) (Buf->curr >= Buf->len)
#endif
