#ifndef __GTK_EDIT_H__
#define __GTK_EDIT_H__

#include <gdk/gdk.h>
#include <gtk/gtkwidget.h>

#ifdef __cplusplus
extern "C" {
#endif				/* __cplusplus */

#define GTK_EDIT(obj)          GTK_CHECK_CAST (obj, gtk_edit_get_type (), GtkEdit)
#define GTK_EDIT_CLASS(klass)  GTK_CHECK_CLASS_CAST (klass, gtk_edit_get_type (), GtkEditClass)
#define GTK_IS_EDIT(obj)       GTK_CHECK_TYPE (obj, gtk_edit_get_type ())

    typedef struct _GtkEdit GtkEdit;
    typedef struct _GtkEditClass GtkEditClass;

/* this is the line that will port all the code to Gtk */
#define WEdit GtkEdit

#define SEARCH_DIALOG_OPTION_NO_SCANF	1
#define SEARCH_DIALOG_OPTION_NO_REGEX	2
#define SEARCH_DIALOG_OPTION_NO_CASE	4
#define SEARCH_DIALOG_OPTION_BACKWARDS	8

#define CLIP_FILE "/.cedit/cooledit.clip"	/* FIXME: need to be configurable */
#define MACRO_FILE "/.cedit/cooledit.macros"
#define BLOCK_FILE "/.cedit/cooledit.block"
#define ERROR_FILE "/.cedit/cooledit.error"
#define TEMP_FILE "/.cedit/cooledit.temp"
#define SCRIPT_FILE "/.cedit/cooledit.script"
#define EDIT_DIR "/.cedit"

#define EDIT_KEY_EMULATION_NORMAL 0
#define EDIT_KEY_EMULATION_EMACS  1

#define REDRAW_LINE          (1 << 0)
#define REDRAW_LINE_ABOVE    (1 << 1)
#define REDRAW_LINE_BELOW    (1 << 2)
#define REDRAW_AFTER_CURSOR  (1 << 3)
#define REDRAW_BEFORE_CURSOR (1 << 4)
#define REDRAW_PAGE          (1 << 5)
#define REDRAW_IN_BOUNDS     (1 << 6)
#define REDRAW_CHAR_ONLY     (1 << 7)
#define REDRAW_COMPLETELY    (1 << 8)

#define MOD_ABNORMAL		(1 << 0)
#define MOD_UNDERLINED		(1 << 1)
#define MOD_BOLD		(1 << 2)
#define MOD_HIGHLIGHTED		(1 << 3)
#define MOD_MARKED		(1 << 4)
#define MOD_ITALIC		(1 << 5)
#define MOD_CURSOR		(1 << 6)
#define MOD_INVERSE		(1 << 7)

#define EDIT_TEXT_HORIZONTAL_OFFSET 4
#define EDIT_TEXT_VERTICAL_OFFSET 3

#define EDIT_RIGHT_EXTREME option_edit_right_extreme
#define EDIT_LEFT_EXTREME option_edit_left_extreme
#define EDIT_TOP_EXTREME option_edit_top_extreme
#define EDIT_BOTTOM_EXTREME option_edit_bottom_extreme

#define MAX_MACRO_LENGTH 1024

/* there are a maximum of ... */
#define MAXBUFF 1024
/* ... edit buffers, each of which is ... */
#define EDIT_BUF_SIZE 16384
/* ...bytes in size. */

/* x / EDIT_BUF_SIZE equals x >> ... */
#define S_EDIT_BUF_SIZE 14

/* x % EDIT_BUF_SIZE is equal to x && ... */
#define M_EDIT_BUF_SIZE 16383

#define SIZE_LIMIT (EDIT_BUF_SIZE * (MAXBUFF - 2))
/* Note a 16k stack is 64k of data and enough to hold (usually) around 10
   pages of undo info. */

/* undo stack */
#define START_STACK_SIZE 32


/* some codes that may be pushed onto or returned from the undo stack: */
#define CURS_LEFT 601
#define CURS_RIGHT 602
#define DELETE 603
#define BACKSPACE 604
#define STACK_BOTTOM 605
#define CURS_LEFT_LOTS 606
#define CURS_RIGHT_LOTS 607
#define MARK_1 1000
#define MARK_2 700000000
#define KEY_PRESS 1400000000

/* Tabs spaces: (sofar only HALF_TAB_SIZE is used: */
#define TAB_SIZE		option_tab_spacing
#define HALF_TAB_SIZE		((int) option_tab_spacing / 2)

    struct macro {
	short command;
	short ch;
    };

    struct selection {
	unsigned char *text;
	int len;
    };

    struct _GtkEditClass {
	GtkWidgetClass parent_class;
    };

    struct _GtkEdit {
	GtkWidget widget;

	int from_here;

/* columns and lines of text */
	int num_widget_lines;
	int num_widget_columns;
	int stopped;

	char *filename;		/* Name of the file */
	char *dir;		/* current directory */

/* dynamic buffers and curser position for editor: */
	long curs1;		/* position of the cursor from the beginning of the file. */
	long curs2;		/* position from the end of the file */
	unsigned char *buffers1[MAXBUFF + 1];	/* all data up to curs1 */
	unsigned char *buffers2[MAXBUFF + 1];	/* all data from end of file down to curs2 */

/* search variables */
	long search_start;	/* First character to start searching from */
	int found_len;		/* Length of found string or 0 if none was found */
	long found_start;	/* the found word from a search - start position */

/* display information */
	long last_byte;		/* Last byte of file */
	long start_display;	/* First char displayed */
	long start_col;		/* First displayed column, negative */
	long curs_row;		/* row position of curser on the screen */
	long curs_col;		/* column position on screen */
	int force;		/* how much of the screen do we redraw? */
	unsigned char overwrite;
	unsigned char modified;	/* has the file been changed?: 1 if char inserted or
				   deleted at all since last load or save */
	unsigned char highlight;
	long prev_col;		/* recent column position of the curser - used when moving
				   up or down past lines that are shorter than the current line */
	long curs_line;		/* line number of the cursor. */
	long start_line;	/* line nummber of the top of the page */

/* file info */
	long total_lines;	/* total lines in the file */
	long mark1;		/* position of highlight start */
	long mark2;		/* position of highlight end */
	int column1;		/* position of column highlight start */
	int column2;		/* position of column highlight end */
	long bracket;		/* position of a matching bracket */

/* undo stack and pointers */
	unsigned long stack_pointer;
	long *undo_stack;
	unsigned long stack_size;
	unsigned long stack_size_mask;
	unsigned long stack_bottom;
	struct stat stat;

	int to_here;		/* dummy marker */

/* macro stuff */
	int macro_i;		/* -1 if not recording index to macro[] otherwise */
	struct macro macro[MAX_MACRO_LENGTH];
    };

    int edit_get_byte (GtkEdit * edit, long byte_index);
    void edit_render_expose (GtkEdit * edit, XExposeEvent * xexpose);
    void edit_render_tidbits (struct cool_widget *w);
    int eh_editor (CWidget * w, XEvent * xevent, CEvent * cwevent);
    void edit_draw_menus (Window parent, int x, int y);
    void edit_run_make (void);
    void edit_change_directory (void);
    int edit_man_page_cmd (GtkEdit * edit);
    void edit_search_replace_dialog (Window parent, int x, int y, char **search_text, char **replace_text, char **arg_order, char *heading, int option);
    void edit_search_dialog (GtkEdit * edit, char **search_text);
    long edit_find (long search_start, unsigned char *exp, int *len, long last_byte, int (*get_byte) (void *, long), void *data);
    void edit_set_foreground_colors (unsigned long normal, unsigned long bold, unsigned long italic);
    void edit_set_background_colors (unsigned long normal, unsigned long abnormal, unsigned long marked, unsigned long marked_abnormal, unsigned long highlighted);
    void edit_set_cursor_color (unsigned long c);
    void CRefreshEditor (GtkEdit * edit);
    void edit_set_user_command (void (*func) (GtkEdit *, int));
    void edit_draw_this_line_proportional (GtkEdit * edit, long b, int curs_row, int start_column, int end_column);
    unsigned char get_international_character (unsigned char key_press);
    void edit_set_user_key_function (int (*user_def_key_func) (unsigned int, unsigned int, KeySym keysym));
    int edit_load_file (GtkEdit * edit, const char *filename, const char *text);
    int edit_count_lines (GtkEdit * edit, long current, int upto);
    long edit_move_forward (GtkEdit * edit, long current, int lines, long upto);
    long edit_move_forward3 (GtkEdit * edit, long current, int cols, long upto);
    long edit_move_backward (GtkEdit * edit, long current, int lines);
    void edit_scroll_screen_over_cursor (GtkEdit * edit);
    void edit_render_keypress (GtkEdit * edit);
    void edit_scroll_upward (GtkEdit * edit, unsigned long i);
    void edit_scroll_downward (GtkEdit * edit, int i);
    void edit_scroll_right (GtkEdit * edit, int i);
    void edit_scroll_left (GtkEdit * edit, int i);
    int edit_get_col (GtkEdit * edit);
    long edit_bol (GtkEdit * edit, long current);
    long edit_eol (GtkEdit * edit, long current);
    void edit_update_curs_row (GtkEdit * edit);
    void edit_update_curs_col (GtkEdit * edit);
    void edit_block_copy_cmd (GtkEdit * edit);
    void edit_block_move_cmd (GtkEdit * edit);
    int edit_block_delete_cmd (GtkEdit * edit);
    int edit_delete (GtkEdit * edit);
    void edit_insert (GtkEdit * edit, int c);
    int edit_cursor_move (GtkEdit * edit, long increment);
    void edit_push_action (GtkEdit * edit, long c,...);
    void edit_push_key_press (GtkEdit * edit);
    void edit_insert_ahead (GtkEdit * edit, int c);
    int edit_save_file (GtkEdit * edit, const char *filename);
    int edit_save_cmd (GtkEdit * edit);
    int edit_save_confirm_cmd (GtkEdit * edit);
    int edit_save_as_cmd (GtkEdit * edit);
    GtkEdit *edit_init (GtkEdit * e, int lines, int columns, const char *filename, const char *text, const char *dir);
    int edit_clean (GtkEdit * edit);
    int edit_renew (GtkEdit * edit);
    int edit_new_cmd (GtkEdit * edit);
    int edit_reload (GtkEdit * edit, const char *filename, const char *text, const char *dir);
    int edit_load_cmd (GtkEdit * edit);
    void edit_mark_cmd (GtkEdit * edit, int unmark);
    void edit_set_markers (GtkEdit * edit, long m1, long m2, int c1, int c2);
    void edit_push_markers (GtkEdit * edit);
    void edit_quit_cmd (GtkEdit * edit);
    void edit_replace_cmd (GtkEdit * edit, int again);
    void edit_search_cmd (GtkEdit * edit, int again);
    int edit_save_block_cmd (GtkEdit * edit);
    int edit_insert_file_cmd (GtkEdit * edit);
    int edit_insert_file (GtkEdit * edit, const char *filename);
    void edit_block_process_cmd (GtkEdit * edit, const char *shell_cmd, int block);
    char *catstrs (const char *first,...);
    void edit_refresh_cmd (GtkEdit * edit);
    void edit_date_cmd (GtkEdit * edit);
    void edit_goto_cmd (GtkEdit * edit);
    int eval_marks (GtkEdit * edit, long *start_mark, long *end_mark);
    void edit_status (GtkEdit * edit);
    int edit_execute_cmd (GtkEdit * edit, int command, int char_for_insertion);
    int edit_execute_key_command (GtkEdit * edit, int command, int char_for_insertion);
    void edit_update_screen (GtkEdit * edit);
    int edit_printf (GtkEdit * e, const char *fmt,...);
    int edit_print_string (GtkEdit * e, const char *s);
    void edit_move_to_line (GtkEdit * e, long line);
    void edit_move_display (GtkEdit * e, long line);
    void edit_word_wrap (GtkEdit * edit);
    unsigned char *edit_get_block (GtkEdit * edit, long start, long finish, int *l);
    int edit_sort_cmd (GtkEdit * edit);
    void edit_help_cmd (GtkEdit * edit);
    void edit_left_word_move (GtkEdit * edit);
    void edit_right_word_move (GtkEdit * edit);
    void edit_get_selection (GtkEdit * edit);
    int edit_save_macro_cmd (GtkEdit * edit, struct macro macro[], int n);
    int edit_load_macro_cmd (GtkEdit * edit, struct macro macro[], int *n, int k);
    void edit_delete_macro_cmd (GtkEdit * edit);
    int edit_copy_to_X_buf_cmd (GtkEdit * edit);
    int edit_cut_to_X_buf_cmd (GtkEdit * edit);
    void edit_paste_from_X_buf_cmd (GtkEdit * edit);
    void edit_paste_from_history (GtkEdit * edit);
    void edit_split_filename (GtkEdit * edit, char *name);

#define MY_O_TEXT 0
#define WIN_MESSAGES edit->widget->mainid, 20, 20

#define edit_get_load_file(d,f,h) CGetLoadFile(WIN_MESSAGES,d,f,h)
#define edit_get_save_file(d,f,h) CGetSaveFile(WIN_MESSAGES,d,f,h)
#define edit_error_dialog(h,t) CErrorDialog(WIN_MESSAGES,h,"%s",t)
#define edit_message_dialog(h,t) CMessageDialog(WIN_MESSAGES,0,h,"%s",t)
#define edit_query_dialog2(h,t,a,b) CQueryDialog(WIN_MESSAGES,h,t,a,b,0)
#define edit_query_dialog3(h,t,a,b,c) CQueryDialog(WIN_MESSAGES,h,t,a,b,c,0)
#define edit_query_dialog4(h,t,a,b,c,d) CQueryDialog(WIN_MESSAGES,h,t,a,b,c,d,0)

    extern char *home_dir;

#define NUM_SELECTION_HISTORY 32

#ifdef _EDIT_C

    struct selection selection =
    {0, 0};
    int current_selection = 0;
/* Note: selection.text = selection_history[current_selection].text */
    struct selection selection_history[NUM_SELECTION_HISTORY] =
    {
	{0, 0},
	{0, 0},
	{0, 0},
	{0, 0},
	{0, 0},
	{0, 0},
	{0, 0},
	{0, 0},
	{0, 0},
	{0, 0},
    };

    int option_word_wrap_line_length = 0;
    int option_international_characters = 0;
    int option_tab_spacing = 8;
    int option_fill_tabs_with_spaces = 0;
    int option_return_does_auto_indent = 1;
    int option_backspace_through_tabs = 0;
    int option_fake_half_tabs = 1;
    int option_save_mode = 0;
    int option_backup_ext_int = -1;
    int option_find_bracket = 1;
    int option_max_undo = 8192;

    int option_editor_fg_normal = 26;
    int option_editor_fg_bold = 8;
    int option_editor_fg_italic = 10;

    int option_edit_right_extreme = 0;
    int option_edit_left_extreme = 0;
    int option_edit_top_extreme = 0;
    int option_edit_bottom_extreme = 0;

    int option_editor_bg_normal = 1;
    int option_editor_bg_abnormal = 0;
    int option_editor_bg_marked = 2;
    int option_editor_bg_marked_abnormal = 9;
    int option_editor_bg_highlighted = 12;
    int option_editor_fg_cursor = 18;

    char *option_whole_chars_search = "0123456789abcdefghijklmnopqrstuvwxyz_";
    char *option_whole_chars_move = "0123456789abcdefghijklmnopqrstuvwxyz_; ,[](){}";
    char *option_backup_ext = "~";

#else				/* ! _EDIT_C */

    extern struct selection selection;
    extern struct selection selection_history[];
    extern int current_selection;
    extern int option_word_wrap_line_length;
    extern int option_international_characters;
    extern int option_tab_spacing;
    extern int option_fill_tabs_with_spaces;
    extern int option_return_does_auto_indent;
    extern int option_backspace_through_tabs;
    extern int option_fake_half_tabs;
    extern int option_save_mode;
    extern int option_backup_ext_int;
    extern int option_find_bracket;
    extern int option_max_undo;

    extern int option_editor_fg_normal;
    extern int option_editor_fg_bold;
    extern int option_editor_fg_italic;

    extern int option_edit_right_extreme;
    extern int option_edit_left_extreme;
    extern int option_edit_top_extreme;
    extern int option_edit_bottom_extreme;

    extern int option_editor_bg_normal;
    extern int option_editor_bg_abnormal;
    extern int option_editor_bg_marked;
    extern int option_editor_bg_marked_abnormal;
    extern int option_editor_bg_highlighted;
    extern int option_editor_fg_cursor;

    extern char *option_whole_chars_search;
    extern char *option_whole_chars_move;
    extern char *option_backup_ext;

#endif				/* ! _EDIT_C */

#ifdef __cplusplus
}

#endif				/*__cplusplus */
#endif				/*__GTK_EDIT_H__ */
