#include <config.h>
#include "coolwidget.h"

int just_dropped_something = 0;

#include "edit.h"
#include <X11/Xmd.h>
#include <X11/Xatom.h>
#include "app_glob.c"
#include "coollocal.h"
#include "editcmddef.h"

void mouse_mark (
    void  *data,			/* generic structure */
    XEvent * event,			/* XEvent */
    CEvent * ce,			/* CWEvent */
    void (*xy) (int, int, int *, int *),
					/* (row, pixel-x-pos) from click pos (x,y) */
    long (*cp) (void *, int, int),	/* get buffer offset from (row, pixel-x-pos) */
    int (*marks) (void *, long *, long *),
					/* get mark start and end */
    int (*range) (void *, long, long, long),
					/* is click between marks? */
    void (*fin_mark) (void *),		/* set marks to not-moving mode (optional) */
    void (*move_mark) (void *),		/* set marks to moving mode (optional) */
    void (*release_mark) (void *, XEvent *),
					/* set marks when button is release.
					   possibly sets selection owner (could be
					   same as fin_mark) (optional) */
    char * (*get_block) (void *, long, long, int *, int *),
					/* return block from between marks,
					   result is free'd by mouse_mark
					   int *, int * is Dnd type, length */
    void (*move) (void *, long, int),	/* move cursor, int is row if needed */
    void (*motion) (void *, long),	/* move second mark position */
    void (*dclick) (void *, XEvent *),	/* called on double click (optional) */
    void (*redraw) (void *, long)	/* called before exit, passing mouse
					   pos in buffer (optional) */
)
{
    static Window win_press = 0;
    static int state = 0;	/* 0 = button up, 1 = button pressed, 2 = button pressed and dragging */
    static int x_last, y_last;
    long click;

    if (event->type == ButtonPress) {
	long start_mark, end_mark;
	state = 1;
	win_press = event->xbutton.window;
	(*xy) (event->xbutton.x, event->xbutton.y, &x_last, &y_last);
	click = (*cp) (data, x_last, y_last);
	if (!(*marks) (data, &start_mark, &end_mark)) {
	    if ((*range) (data, start_mark, end_mark, click)) {		/* if clicked on highlighted text */
		unsigned char *t;
		int type;
		int l;
		t = (unsigned char *) (*get_block) (data, start_mark, end_mark, &type, &l);
		if (t) {
		    just_dropped_something = 1;
		    CDrag (event->xbutton.window, type, t, l, event->xbutton.button == Button1 ? Button1Mask : 0);
		    free (t);
		}
		if (fin_mark)
		    (*fin_mark) (data);
		return;
	    }
	}
	just_dropped_something = 0;
	if (fin_mark)
	    (*fin_mark) (data);
	(*move) (data, click, y_last);
	if (ce->double_click && dclick) {
	    (*dclick) (data, event);
	    state = 0;
	}
	if (redraw)
	    (*redraw) (data, click);
    } else if (event->type == ButtonRelease && state > 0 && win_press == event->xbutton.window && !ce->double_click) {
	int x, y;
	long start_mark, end_mark;
	(*xy) (event->xbutton.x, event->xbutton.y, &x, &y);
	click = (*cp) (data, x, y);
	(*move) (data, click, y);
	if (state == 2)
	    goto unhighlight;
	if (!(*marks) (data, &start_mark, &end_mark)) {
	    if ((*range) (data, start_mark, end_mark, click)) {		/* if clicked on highlighted text */
	      unhighlight:
		if (release_mark)
		    (*release_mark) (data, event);
	    }
	}
	state = 0;
	if (redraw)
	    (*redraw) (data, click);
    } else if (event->type == MotionNotify && state > 0 && win_press == event->xbutton.window && event->xbutton.state) {
	int x, y;
	if (!event->xmotion.state)
	    return;
	(*xy) (event->xbutton.x, event->xbutton.y, &x, &y);
	if (x == x_last && y == y_last && state == 1)
	    return;
	click = (*cp) (data, x, y);
	if (state == 1) {
	    state = 2;
	    if (move_mark)
		(move_mark) (data);
	}
	(*move) (data, click, y);
	if (motion)
	    (*motion) (data, click);
	if (redraw)
	    (*redraw) (data, click);
    }
}

