/* pixmap.c - creates XImages and pixmaps from XPM formats
   Copyright (C) 1996, 1997 Paul Sheer

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  
 */

#include "coolwidget.h"


XImage *CCreateImage (const char *data[], int width, int height, char start_char)
{
    XImage *image;
    int bpp, x, y, u;

    if (CDepth <= 8) {
	bpp = 1;
    } else if (CDepth <= 16) {
	bpp = 2;
    } else {
	bpp = 4;
    }

    if (width & 1)
	u = 8;
    else if (width & 2)
	u = 16;
    else
	u = 32;
    
    image = XCreateImage (CDisplay, CVisual, CDepth, ZPixmap,
    		0, CMalloc (width * height * bpp), width, height, u, 0);
    if (!image)
	return 0;

    for (y = 0; y < height; y++)
	for (x = 0; x < width; x++) {
	    if (data[y][x] == ' ')
		XPutPixel (image, x, y, COLOR_FLAT);
	    else
		XPutPixel (image, x, y, color_pixels[data[y][x] - start_char]);
	}
    return image;
}


Pixmap CCreatePixmap (const char *data[], int width, int height, char start_char)
{
    Pixmap pixmap;
    XImage *image;
    image = CCreateImage (data, width, height, start_char);
    if (!image)
	return 0;
    pixmap = XCreatePixmap (CDisplay, CRoot, width, height, CDepth);
    XPutImage (CDisplay, pixmap, CGC, image, 0, 0, 0, 0, width, height);
    free (image->data);
    image->data = 0;
    XDestroyImage (image);
    return pixmap;
}

