
#define JXDND_DEBUG_MSGS	0	/* boolean */
#define JXDND_SOURCE_DELAY	0	/* time in seconds */
#define JXDND_TARGET_DELAY	0	/* time in seconds */

const Atom kCurrentDNDVersion = 0;

const int kWaitForLastStatusTime = 10 * CLOCKS_PER_SEC;

/* atom names */

#define kDNDSelectionXAtomName "XdndSelection";

#define kDNDAwareXAtomName "XdndAware";
#define kDNDTypeListXAtomName "XdndTypeList"
#define kDNDEnterXAtomName "XdndEnter"
#define kDNDHereXAtomName "XdndPosition"
#define kDNDStatusXAtomName "XdndStatus"
#define kDNDLeaveXAtomName "XdndLeave"
#define kDNDDropXAtomName "XdndDrop"

/* message data */

enum {
    kDNDEnterTypeCount = 3,	// max number of types in XdndEnter message

/* DNDEnter */

    kDNDEnterWindow = 0,	// source window (sender)
     kDNDEnterFlags,		// version in high byte, bit 0 => more data types
     kDNDEnterType1,		// first available data type
     kDNDEnterType2,		// second available data type
     kDNDEnterType3,		// third available data type

/* DNDEnter flags */

    kDNDEnterMoreTypesFlag = 1,	// set if there are more than kDNDEnterTypeCount
     kDNDEnterVersionRShift = 24,	// right shift to position version number

/* DNDHere */

    kDNDHereWindow = 0,		// source window (sender)
     kDNDHereFlags,		// reserved
     kDNDHerePt,		// x,y coords of mouse (root window coords)

/* DNDStatus */

    kDNDStatusWindow = 0,	// target window (sender)
     kDNDStatusFlags,		// flags returned by target
     kDNDStatusPt,		// x,y of "no msg" rectangle (root window coords)
     kDNDStatusArea,		// w,h of "no msg" rectangle

/* DNDStatus flags */

    kDNDStatusAcceptDropFlag = 1,	// set if target will accept the drop
     kDNDStatusSendHereFlag = 2,	// set if target wants stream of XdndPosition

/* DNDLeave */

    kDNDLeaveWindow = 0,	// source window (sender)
     kDNDLeaveFlags,		// reserved

/* DNDDrop */

    kDNDDropWindow = 0,		// source window (sender)
     kDNDDropFlags		// reserved
};

/* default cursor */

#define kDefaultDNDCursorName "JXDNDCursor"

static char dnd_cursor_bits[] =
{
  0x00, 0x00, 0x00, 0xfe, 0xff, 0x0f, 0x02, 0x00, 0x08, 0x02, 0x00, 0x08,
  0x02, 0x00, 0x08, 0x02, 0x00, 0x08, 0x02, 0x00, 0x08, 0x02, 0x00, 0x08,
  0x02, 0x00, 0x08, 0x02, 0x00, 0x08, 0x02, 0x04, 0x08, 0x02, 0x0c, 0x08,
  0x02, 0x1c, 0x08, 0x02, 0x3c, 0x08, 0x02, 0x7c, 0x08, 0x02, 0xfc, 0x08,
  0x02, 0xfc, 0x09, 0x02, 0xfc, 0x0b, 0x02, 0x7c, 0x08, 0xfe, 0x6d, 0x0f,
  0x00, 0xc4, 0x00, 0x00, 0xc0, 0x00, 0x00, 0x80, 0x01, 0x00, 0x80, 0x01,
    0x00, 0x00, 0x00};

static char dnd_mask_bits[] =
{
  0xff, 0xff, 0x1f, 0xff, 0xff, 0x1f, 0xff, 0xff, 0x1f, 0x07, 0x00, 0x1c,
  0x07, 0x00, 0x1c, 0x07, 0x00, 0x1c, 0x07, 0x00, 0x1c, 0x07, 0x00, 0x1c,
  0x07, 0x00, 0x1c, 0x07, 0x06, 0x1c, 0x07, 0x0e, 0x1c, 0x07, 0x1e, 0x1c,
  0x07, 0x3e, 0x1c, 0x07, 0x7e, 0x1c, 0x07, 0xfe, 0x1c, 0x07, 0xfe, 0x1d,
  0x07, 0xfe, 0x1f, 0x07, 0xfe, 0x1f, 0xff, 0xff, 0x1f, 0xff, 0xff, 0x1e,
  0xff, 0xef, 0x1f, 0x00, 0xe6, 0x01, 0x00, 0xc0, 0x03, 0x00, 0xc0, 0x03,
    0x00, 0x80, 0x01};

typedef struct {
    int Width, Height;
    unsigned char *ImageData, *MaskData;
    int HotSpotX, HotSpotY;
    Pixmap ImagePixmap, MaskPixmap;
    Cursor CursorID;
} CursorData;

static CursorData kDefaultDNDCursor =
{
    21, 25, dnd_cursor_bits, dnd_mask_bits, 10, 10, 0, 0, 0
};

Display *itsDisplay = 0;
CWidget *itsDragger;		/* widget initiating drag; not owned */
Window itsDraggerWindow;	/* window of itsDragger; not owned */
Atom itsDraggerTypeList[256];	/* data types supported by itsDragger */
int itsDraggerTypeListLength;	/* number of data types supported in itsDraggerTypeList (Up to 256) */

Atom itsDNDVersion;		/* version being used */
Window itsMouseWindow		/* window that mouse is in; not owned */
CWidget *itsMouseContainer;	/* widget that mouse is in; NULL if inter-app; not owned */

int itsWillAcceptDropFlag;	/* True if target will accept drop */
Bool itsWaitForStatusFlag;	/* True if waiting for XdndStatus message */
Bool itsSendHereMsgFlag;	/* True if need another XdndPosition with itsPrevHandleDNDPt */
Bool itsReceivedStatusFlag;	/* True if received any XdndStatus from current target */
Bool itsUseMouseRectFlag;	/* True if use itsMouseRectR */
Uknown[JRect] itsMouseRectR;	/* don't send another XdndPosition while inside here */

XPoint itsPrevMousePt;		/* last XdndPosition coordinates (local coords of target) */
XPoint itsPrevHandleDNDPt;	/* last JPoint sent to HandleDND() */

Cursor itsDefDNDCursor;

Atom itsDNDSelectionName;
Atom itsDNDAwareXAtom;
Atom itsDNDTypeListXAtom;
Atom itsDNDEnterXAtom;
Atom itsDNDHereXAtom;
Atom itsDNDStatusXAtom;
Atom itsDNDLeaveXAtom;
Atom itsDNDDropXAtom;

/******************************************************************************
 Constructor

 ******************************************************************************/

void JXDNDManager (void)
{
    itsDisplay = CDisplay;
    itsDragger = NULL;
    itsDraggerWindow = None;
    itsMouseWindow = None;
    itsMouseContainer = NULL;
    Colormap colormap;
    Window root;
    XColor Black, White;
    Pixmap pixmap, mask;

    colormap = DefaultColormap (CDisplay, DefaultScreen (CDisplay));
    root = DefaultRootWindow (CDisplay);

    Black.pixel = BlackPixel (CDisplay, DefaultScreen (CDisplay));
    White.pixel = WhitePixel (CDisplay, DefaultScreen (CDisplay));
    XQueryColor (CDisplay, colormap, &Black);
    XQueryColor (CDisplay, colormap, &White);

    kDefaultDNDCursor.ImagePixmap =
	XCreateBitmapFromData (CDisplay, root,
			       (char *) kDefaultDNDCursor.ImageData,
			       kDefaultDNDCursor.Width,
			       kDefaultDNDCursor.Height);
    kDefaultDNDCursor.MaskPixmap =
	XCreateBitmapFromData (CDisplay, root,
			       (char *) kDefaultDNDCursor.MaskData,
			       kDefaultDNDCursor.Width,
			       kDefaultDNDCursor.Height);
    kDefaultDNDCursor.CursorID =
	XCreatePixmapCursor (CDisplay, kDefaultDNDCursor.ImagePixmap,
			     kDefaultDNDCursor.MaskPixmap,
			     &Black, &White,
			     kDefaultDNDCursor.HotSpotX,
			     kDefaultDNDCursor.HotSpotY);

    itsDefDNDCursor = kDefaultDNDCursor.CursorID;

    /* create required X atoms */

    itsDNDSelectionName = XInternAtom (CDisplay, kDNDSelectionXAtomName, False);

    itsDNDAwareXAtom = XInternAtom (CDisplay, kDNDAwareXAtomName, False);
    itsDNDTypeListXAtom = XInternAtom (CDisplay, kDNDTypeListXAtomName, False);

    itsDNDEnterXAtom = XInternAtom (CDisplay, kDNDEnterXAtomName, False);
    itsDNDHereXAtom = XInternAtom (CDisplay, kDNDHereXAtomName, False);
    itsDNDStatusXAtom = XInternAtom (CDisplay, kDNDStatusXAtomName, False);
    itsDNDLeaveXAtom = XInternAtom (CDisplay, kDNDLeaveXAtomName, False);
    itsDNDDropXAtom = XInternAtom (CDisplay, kDNDDropXAtomName, False);
}

void JXDNDManager (void)
{
/* what ever we need to destroy stuff */
    return;
}


/******************************************************************************
 BeginDND

	Returns False if we are unable to initiate DND.

 ******************************************************************************/

Bool BeginDND (CWidget * widget, XPoint pt, int buttonStates, int modifiers)
{
    if (itsDragger != NULL)
	return False;
    itsMouseWindow = None;
    itsMouseContainer = NULL;
    if (!GetSelectionTargetTypes (itsDraggerTypeList, itsDraggerTypeListLength))
	return False;
    if (!BecomeSelectionOwner (itsDNDSelectionName))
	return False;
    itsDragger = widget;
    itsDraggerWindow = widget->winid;
    BecomeDNDSource (itsDragger);
    XGrabPointer (CDisplay, DefaultRootWindow (CDisplay), False,
		  ButtonMotionMask | ButtonPressMask | ButtonReleaseMask | LeaveWindowMask | EnterWindowMask,
	       GrabModeSync, GrabModeAsync, DefaultRootWindow (CDisplay),
		  itsDefDNDCursor, CurrentTime);
    DNDInit (pt, buttonStates, modifiers);
    AnnounceTypeList (itsDraggerWindow, itsDraggerTypeList, itsDraggerTypeListLength);
    HandleDND (pt);
    return True;
}

/******************************************************************************
 HandleDND

 ******************************************************************************/

void HandleDND (XPoint pt)
{
    Window root;
    JXContainer *dropWidget;
    Window xWindow;
    JXWindow *w;
    Window root_return;
    Window child_return;
    int root_x_return, root_y_return, win_x_return, win_y_return;

    if (!itsDragger)
	return;

    root = DefaultRootwindow (CDisplay);

    if (!XQueryPointer (CDisplay, root, root_return, child_return,
	    &root_x_return, &root_y_return, &win_x_return, &win_y_return,
			&mask_return))
	return;

    if (!XQueryPointer (CDisplay, root, root_return, child_return,
	    &root_x_return, &root_y_return, &win_x_return, &win_y_return,
			&mask_return))
	return;

    root = DefaultRootWindow (CDisplay);

    /* If a local window is deactivated, dropWidget will be NULL. */
    /* To avoid sending ourselves client messages, we have to treat */
    /* this case as the root window. */

    if (dropWidget == NULL && xWindow != itsMouseWindow &&
	itsDisplay->FindWindow (xWindow, &w)) {
	xWindow = rootWindow;
    }
    /* treat windows without XdndAware as the root window */

    if (dropWidget == NULL &&
	xWindow != itsMouseWindow && xWindow != rootWindow &&
	!IsDNDAware (xWindow, &itsDNDVersion)) {
	xWindow = rootWindow;
    }
    Bool forceSendDNDHere = False;
    if (itsMouseContainer == itsDragger && itsWillAcceptDropFlag &&
	((xWindow == itsDraggerWindow && dropWidget != itsDragger &&
	  (dropWidget == NULL ||
	   !dropWidget->WillAcceptDrop (itsDraggerTypeList, itsDraggerTypeListLength, False))) ||
	 xWindow == rootWindow)) {
	/* keep sending messages to itsDragger */
	/* (checking dropWidget != itsDragger insures that WillAcceptDrop() */
	/*  isn't continuously called on itsDragger) */
    } else if (xWindow != itsMouseWindow || dropWidget != itsMouseContainer) {
	SendDNDLeave ();
	SendDNDEnter (xWindow, dropWidget);
	forceSendDNDHere = True;
    }
    if (forceSendDNDHere || pt != itsPrevHandleDNDPt) {
	itsPrevHandleDNDPt = pt;
	SendDNDHere (pt);
    }
    /* slow to receive next mouse event */

#if JXDND_SOURCE_DELAY > 0
    JWait (JXDND_SOURCE_DELAY);
#endif
}

/******************************************************************************
 FinishDND

 ******************************************************************************/

void FinishDND (void)
{
    if (itsDragger == NULL)
	return;

    if (WaitForLastStatusMsg ()) {
	SendDNDDrop ();
    } else {
	SendDNDLeave ();
    }

    itsDragger = NULL;
    itsDraggerWindow = None;
    itsMouseWindow = None;
    itsMouseContainer = NULL;
}

/******************************************************************************
 EnableDND

	Creates the XdndAware property on the specified window.

	We cannot include a list of data types because JXWindows contain
	many widgets.

 ******************************************************************************/

void EnableDND (Window xWindow)
{
    XChangeProperty (CDisplay, xWindow, itsDNDAwareXAtom, XA_ATOM, 32,
		     PropModeReplace, (char *) &kCurrentDNDVersion, 1);
}

/******************************************************************************
 IsDNDAware (private)

	Returns True if the given X window supports the XDND protocol.

	*vers is the version to use.

 ******************************************************************************/

Bool IsDNDAware (Window xWindow, int *vers)
{
    *vers = 0;

    Atom actualType;
    int actualFormat;
    unsigned long itemCount, remainingBytes;
    unsigned char *rawData = NULL;
    XGetWindowProperty (CDisplay, xWindow, itsDNDAwareXAtom,
			0, LONG_MAX, False, XA_ATOM,
			&actualType, &actualFormat,
			&itemCount, &remainingBytes, &rawData);

    if (actualType == XA_ATOM && actualFormat == 32 && itemCount > 0) {
	Atom *data = (Atom *) rawData;
	*vers = min (kCurrentDNDVersion, data[0]);

	Bool typesIntersect = True;
	if (itemCount > 1) {
	    typesIntersect = False;

	    const int count = itsDraggerTypeListLength;
	    for (JIndex i = 1; i <= count; i++) {
		const Atom sourceType = itsDraggerTypeList[i];
		for (JIndex j = 1; j < itemCount; j++) {
		    if (data[i] == sourceType) {
			typesIntersect = True;
			break;
		    }
		}
		if (typesIntersect) {
		    break;
		}
	    }
	}
	XFree (rawData);
	return typesIntersect;
    } else {
	XFree (rawData);
	return False;
    }
}

/******************************************************************************
 AnnounceTypeList (private)

	Creates the XdndTypeList property on the specified window if there
	are more than kDNDEnterTypeCount data types.

 ******************************************************************************/

void AnnounceTypeList (Window xWindow, Atom list[], int list_length)
{
    const int typeCount = itsDraggerTypeListLength;
    if (typeCount > kDNDEnterTypeCount) {
	XChangeProperty (CDisplay, xWindow,
			 itsDNDTypeListXAtom, XA_ATOM, 32,
			 PropModeReplace,
			 (char *) list, typeCount);
    }
}

/******************************************************************************


 SendDNDEnter (private)

 ******************************************************************************/ void
void SendDNDEnter (Window xWindow, CWidget *widget) {
    itsMouseWindow = xWindow;
    itsMouseContainer = widget;

    itsWillAcceptDropFlag = False;
    itsWaitForStatusFlag = False;
    itsSendHereMsgFlag = False;
    itsReceivedStatusFlag = False;
    itsUseMouseRectFlag = False;
    itsMouseRectR = JRect (0, 0, 0, 0);

    if (itsMouseContainer != NULL) {
	itsWillAcceptDropFlag =
	    itsMouseContainer->WillAcceptDrop (*itsDraggerTypeList,
		    JConvertToBoolean (itsDragger == itsMouseContainer));
	if (itsWillAcceptDropFlag) {
	    itsMouseContainer->DNDEnter ();
	}
	itsReceivedStatusFlag = True;
    } else if (itsMouseWindow != None &&
	       itsMouseWindow != DefaulrRootWindow (CDisplay)) {
#if JXDND_DEBUG_MSGS
	fprintf ("Sent XdndEnter\n");
#endif

#if JXDND_SOURCE_DELAY > 0
	JWait (JXDND_SOURCE_DELAY);
#endif

	const int typeCount = itsDraggerTypeListLength;

	XEvent xEvent;
	XClientMessageEvent *message;
	message = &xEvent.xclient;

	message->type = ClientMessage;
	message->display = *itsDisplay;
	message->window = itsMouseWindow;
	message->message_type = itsDNDEnterXAtom;
	message->format = 32;

	message->data.l[kDNDEnterWindow] = itsDraggerWindow;

	message->data.l[kDNDEnterFlags] = 0;
	message->data.l[kDNDEnterFlags] |= (itsDNDVersion << kDNDEnterVersionRShift);
	if (typeCount > kDNDEnterTypeCount) {
	    message.data.l[kDNDEnterFlags] |= kDNDEnterMoreTypesFlag;
	}
	message.data.l[kDNDEnterType1] = None;
	message.data.l[kDNDEnterType2] = None;
	message.data.l[kDNDEnterType3] = None;

	const int msgTypeCount = min (typeCount, (int) kDNDEnterTypeCount);
	for (JIndex i = 1; i <= msgTypeCount; i++) {
	    message.data.l[kDNDEnterType1 + i - 1] = itsDraggerTypeList->GetElement (i);
	}

	XSendEvent (CDisplay, itsMouseWindow, 0, 0, &xEvent);
    }
}

/******************************************************************************
 SendDNDHere (private)

 ******************************************************************************/

void SendDNDHere (XPoint pt1)
{
    const JPoint ptG1 = LocalToGlobal (itsDragger->winid, pt1);
    const JPoint ptR = GlobalToRoot (itsDragger->winid, ptG1);

    if (itsMouseContainer != NULL) {
	if (itsWillAcceptDropFlag) {
	    const JPoint ptG2 = RootToGlobal (itsMouseContainer, ptR);
	    itsPrevMousePt = GlobalToLocal (itsMouseContainer, ptG2);
	    DNDHere (itsMouseContainer, itsPrevMousePt, (itsDragger == itsMouseContainer));
	}
    } else if (!itsWaitForStatusFlag &&
	       itsMouseWindow != DefaultRootWindow (CDisplay) &&
	       (!itsUseMouseRectFlag || !itsMouseRectR.Contains (ptR))) {
#if JXDND_DEBUG_MSGS
	printf ("Sent XdndPosition: %d %d\n", ptR.x, ptR.y);
#endif

#if JXDND_SOURCE_DELAY > 0
	JWait (JXDND_SOURCE_DELAY);
#endif

	XEvent xEvent;
	XClientMessageEvent *message = &xEvent.xclient;

	message->type = ClientMessage;
	message->display = *itsDisplay;
	message->window = itsMouseWindow;
	message->message_type = itsDNDHereXAtom;
	message->format = 32;

	message->data.l[kDNDHereWindow] = itsDraggerWindow;
	message->data.l[kDNDHereFlags] = 0;
	message->data.l[kDNDHerePt] = PackPoint (ptR);

	XSendEvent (CDisplay, itsMouseWindow, 0, 0, &xEvent);

	itsWaitForStatusFlag = True;
	itsSendHereMsgFlag = False;
    } else if (itsWaitForStatusFlag) {
	itsSendHereMsgFlag = True;
    }
}
'''
/******************************************************************************
 SendDNDLeave (private)

 ******************************************************************************/

void SendDNDLeave (void)
{
    if (itsMouseContainer != NULL) {
	if (itsWillAcceptDropFlag) {
	    DNDLeave (itsMouseContainer);
	}
    } else if (itsMouseWindow != None &&
	       itsMouseWindow != DefaultRootWindow (CDisplay)) {
#if JXDND_DEBUG_MSGS
	printf ("Sent XdndLeave");
#endif

#if JXDND_SOURCE_DELAY > 0
	JWait (JXDND_SOURCE_DELAY);
#endif

	XEvent xEvent;
	XClientMessageEvent *message = &xEvent.xclient;

	message->type = ClientMessage;
	message->display = *itsDisplay;
	message->window = itsMouseWindow;
	message->message_type = itsDNDLeaveXAtom;
	message->format = 32;

	message->data.l[kDNDLeaveWindow] = itsDraggerWindow;
	message->data.l[kDNDLeaveFlags] = 0;

	XSendEvent (CDisplay, itsMouseWindow, 0, 0, &xEvent);
    }
}

/******************************************************************************
 SendDNDDrop (private)

 ******************************************************************************/

void SendDNDDrop (void)
{
    if (itsMouseContainer != NULL && itsWillAcceptDropFlag) {
	itsMouseContainer->DNDDrop (itsPrevMousePt, *itsDraggerTypeList,
		    JConvertToBoolean (itsDragger == itsMouseContainer));
    } else if (itsMouseWindow != itsDisplay->GetRootWindow () && itsWillAcceptDropFlag) {
#if JXDND_DEBUG_MSGS
	cout << "Sent XdndDrop" << endl;
#endif

#if JXDND_SOURCE_DELAY > 0
	JWait (JXDND_SOURCE_DELAY);
#endif

	XEvent xEvent;
	XClientMessageEvent & message = xEvent.xclient;

	message.type = ClientMessage;
	message.display = *itsDisplay;
	message.window = itsMouseWindow;
	message.message_type = itsDNDDropXAtom;
	message.format = 32;

	message.data.l[kDNDDropWindow] = itsDraggerWindow;
	message.data.l[kDNDDropFlags] = 0;

	itsDisplay->SendXEvent (itsMouseWindow, &xEvent);
    }
}

/******************************************************************************
 SendDNDStatus (private)

	This is sent to let the source know that we are ready for another
	DNDHere message.

 ******************************************************************************/

void
 JXDNDManager::SendDNDStatus
 (
     const Bool willAcceptDrop
) {
    if (itsDraggerWindow != None) {
#if JXDND_DEBUG_MSGS
	cout << "Sent XdndStatus: " << willAcceptDrop << endl;
#endif

#if JXDND_TARGET_DELAY > 0
	JWait (JXDND_TARGET_DELAY);
#endif

	XEvent xEvent;
	XClientMessageEvent & message = xEvent.xclient;

	message.type = ClientMessage;
	message.display = *itsDisplay;
	message.window = itsDraggerWindow;
	message.message_type = itsDNDStatusXAtom;
	message.format = 32;

	message.data.l[kDNDStatusWindow] = itsMouseWindow;
	message.data.l[kDNDStatusFlags] = 0;
	message.data.l[kDNDStatusPt] = 0;	/* empty rectangle */
	message.data.l[kDNDStatusArea] = 0;

	if (willAcceptDrop) {
	    message.data.l[kDNDStatusFlags] |= kDNDStatusAcceptDropFlag;
	}
	itsDisplay->SendXEvent (itsDraggerWindow, &xEvent);
    }
}

/******************************************************************************
 HandleClientMessage 

 ******************************************************************************/

Bool
JXDNDManager: :HandleClientMessage
(
    const XClientMessageEvent & clientMessage
) {
    /* target:  receive and process window-level enter and leave */

    if (clientMessage.message_type == itsDNDEnterXAtom) {
	HandleDNDEnter (clientMessage);
	return True;
    } else if (clientMessage.message_type == itsDNDHereXAtom) {
	HandleDNDHere (clientMessage);
	return True;
    } else if (clientMessage.message_type == itsDNDLeaveXAtom) {
	HandleDNDLeave (clientMessage);
	return True;
    } else if (clientMessage.message_type == itsDNDDropXAtom) {
	HandleDNDDrop (clientMessage);
	return True;
    }
    /* source:  clear our flag when we receive status message */

    else if (clientMessage.message_type == itsDNDStatusXAtom) {
	HandleDNDStatus (clientMessage);
	return True;
    }
    /* not for us */

    else {
	return False;
    }
}

/******************************************************************************
 HandleDNDEnter (private)

 ******************************************************************************/

void
 JXDNDManager::HandleDNDEnter
 (
     const XClientMessageEvent & clientMessage
) {
    assert (clientMessage.message_type == itsDNDEnterXAtom);

    itsDNDVersion = (clientMessage.data.l[kDNDEnterFlags] >> kDNDEnterVersionRShift) & 0xFF;

    if (itsDraggerWindow == None && itsDNDVersion <= kCurrentDNDVersion &&
	clientMessage.data.l[kDNDEnterType1] != None) {
#if JXDND_TARGET_DELAY > 0
#if JXDND_DEBUG_MSGS
	cout << "XdndEnter will arrive" << endl;
#endif
	JWait (JXDND_TARGET_DELAY);
#endif

#if JXDND_DEBUG_MSGS
	cout << "Received XdndEnter" << endl;
#endif

	itsDragger = NULL;
	itsDraggerWindow = clientMessage.data.l[kDNDEnterWindow];
	itsMouseWindow = clientMessage.window;
	itsMouseContainer = NULL;

	XSelectInput (*itsDisplay, itsDraggerWindow, StructureNotifyMask);
	ListenTo (itsDisplay);

	itsDraggerTypeList->RemoveAll ();
	if (clientMessage.data.l[kDNDEnterFlags] & kDNDEnterMoreTypesFlag != 0) {
	    Atom actualType;
	    int actualFormat;
	    unsigned long itemCount, remainingBytes;
	    unsigned char *rawData = NULL;
	    XGetWindowProperty (*itsDisplay, itsDraggerWindow, itsDNDTypeListXAtom,
				0, LONG_MAX, False, XA_ATOM,
				&actualType, &actualFormat,
				&itemCount, &remainingBytes, &rawData);

	    if (actualType == XA_ATOM && actualFormat == 32 && itemCount > 0) {
		Atom *data = static_cast (Atom *, rawData);
		for (JIndex i = 0; i < itemCount; i++) {
		    itsDraggerTypeList->AppendElement (data[i]);
		}
	    }
	    XFree (rawData);
	}
	if (itsDraggerTypeList->IsEmpty ()) {	/* in case window property was empty */
	    for (JIndex i = 1; i <= kDNDEnterTypeCount; i++) {
		const Atom type = clientMessage.data.l[kDNDEnterType1 + i - 1];
		if (type == None) {
		    assert (i > 1);
		    break;
		}
		itsDraggerTypeList->AppendElement (type);
	    }
	}
    }
}

/******************************************************************************
 HandleDNDHere (private)

 ******************************************************************************/

void
 JXDNDManager::HandleDNDHere
 (
     const XClientMessageEvent & clientMessage
) {
    assert (clientMessage.message_type == itsDNDHereXAtom);

    if (itsDraggerWindow != (Window) clientMessage.data.l[kDNDHereWindow]) {
	return;
    }
#if JXDND_TARGET_DELAY > 0
#if JXDND_DEBUG_MSGS
    cout << "XdndPosition will arrive" << endl;
#endif
    JWait (JXDND_TARGET_DELAY);
#endif

#if JXDND_DEBUG_MSGS
    cout << "Received XdndPosition" << endl;
#endif

    const JPoint ptR = UnpackPoint (clientMessage.data.l[kDNDHerePt]);
    JPoint ptG;

    JXWindow *window;
    if (itsMouseContainer != NULL) {
	window = itsMouseContainer->GetWindow ();
	ptG = window->RootToGlobal (ptR);
	JXContainer *newMouseContainer;
	if (itsDisplay->FindMouseContainer (window, ptG, &newMouseContainer) &&
	    newMouseContainer != itsMouseContainer) {
	    if (itsWillAcceptDropFlag) {
		itsMouseContainer->DNDLeave ();
	    }
	    itsMouseContainer = newMouseContainer;
	    itsWillAcceptDropFlag =
		itsMouseContainer->WillAcceptDrop (*itsDraggerTypeList, False);
	    if (itsWillAcceptDropFlag) {
		itsMouseContainer->DNDEnter ();
	    }
	}
    } else if (itsDisplay->FindWindow (clientMessage.window, &window)) {
	ptG = window->RootToGlobal (ptR);
	if (itsDisplay->FindMouseContainer (window, ptG, &itsMouseContainer)) {
	    itsWillAcceptDropFlag =
		itsMouseContainer->WillAcceptDrop (*itsDraggerTypeList, False);
	    if (itsWillAcceptDropFlag) {
		itsMouseContainer->DNDEnter ();
	    }
	}
    }
    if (itsMouseContainer != NULL && itsWillAcceptDropFlag) {
	itsPrevMousePt = itsMouseContainer->GlobalToLocal (ptG);
	itsMouseContainer->DNDHere (itsPrevMousePt, False);
	SendDNDStatus (True);
    } else {
	SendDNDStatus (False);
    }
}

/******************************************************************************
 HandleDNDLeave (private)

 ******************************************************************************/

void
 JXDNDManager::HandleDNDLeave
 (
     const XClientMessageEvent & clientMessage
) {
    assert (clientMessage.message_type == itsDNDLeaveXAtom);

    if (itsDraggerWindow == (Window) clientMessage.data.l[kDNDLeaveWindow]) {
#if JXDND_TARGET_DELAY > 0
#if JXDND_DEBUG_MSGS
	cout << "XdndLeave will arrive" << endl;
#endif
	JWait (JXDND_TARGET_DELAY);
#endif

#if JXDND_DEBUG_MSGS
	cout << "Received XdndLeave" << endl;
#endif

	HandleDNDLeave1 ();
    }
}

/******************************************************************************
 HandleDNDLeave1 (private)

	Called by HandleDNDLeave() and Receive().

 ******************************************************************************/

void
 JXDNDManager::HandleDNDLeave1 ()
{
    if (itsMouseContainer != NULL && itsWillAcceptDropFlag) {
	itsMouseContainer->DNDLeave ();
    }
    HandleDNDFinished ();
}

/******************************************************************************
 HandleDNDDrop (private)

 ******************************************************************************/

void
 JXDNDManager::HandleDNDDrop
 (
     const XClientMessageEvent & clientMessage
) {
    assert (clientMessage.message_type == itsDNDDropXAtom);

    if (itsDraggerWindow == (Window) clientMessage.data.l[kDNDDropWindow]) {
#if JXDND_TARGET_DELAY > 0
#if JXDND_DEBUG_MSGS
	cout << "XdndDrop will arrive" << endl;
#endif
	JWait (JXDND_TARGET_DELAY);
#endif

#if JXDND_DEBUG_MSGS
	cout << "Received XdndDrop" << endl;
#endif

	if (itsMouseContainer != NULL && itsWillAcceptDropFlag) {
	    itsMouseContainer->DNDDrop (itsPrevMousePt, *itsDraggerTypeList, False);
	}
	HandleDNDFinished ();
    }
}

/******************************************************************************
 HandleDNDFinished (private)

	Called by HandleDNDLeave() and HandleDNDDrop().

 ******************************************************************************/

void
 JXDNDManager::HandleDNDFinished ()
{
    XSelectInput (*itsDisplay, itsDraggerWindow, NoEventMask);
    StopListening (itsDisplay);

    itsDragger = NULL;
    itsDraggerWindow = None;
    itsMouseWindow = None;
    itsMouseContainer = NULL;
}

/******************************************************************************
 HandleDNDStatus (private)

 ******************************************************************************/

void
 JXDNDManager::HandleDNDStatus
 (
     const XClientMessageEvent & clientMessage
) {
    assert (clientMessage.message_type == itsDNDStatusXAtom);

    if (itsDragger != NULL &&
     itsMouseWindow == (Window) clientMessage.data.l[kDNDStatusWindow]) {
#if JXDND_SOURCE_DELAY > 0
#if JXDND_DEBUG_MSGS
	cout << "XdndStatus will arrive" << endl;
#endif
	JWait (JXDND_SOURCE_DELAY);
#endif

	itsWaitForStatusFlag = False;
	itsReceivedStatusFlag = True;

	itsWillAcceptDropFlag = JConvertToBoolean (
	clientMessage.data.l[kDNDStatusFlags] & kDNDStatusAcceptDropFlag);

	itsUseMouseRectFlag = JNegate (
	 clientMessage.data.l[kDNDStatusFlags] & kDNDStatusSendHereFlag);

	itsMouseRectR = UnpackRect (clientMessage.data.l[kDNDStatusPt],
				    clientMessage.data.l[kDNDStatusArea]);

#if JXDND_DEBUG_MSGS
	cout << "Received XdndStatus: " << itsWillAcceptDropFlag;
	cout << ' ' << itsUseMouseRectFlag;
	cout << ' ' << itsMouseRectR.top << ' ' << itsMouseRectR.left;
	cout << ' ' << itsMouseRectR.bottom << ' ' << itsMouseRectR.right << endl;
#endif

	if (itsSendHereMsgFlag) {
	    SendDNDHere (itsPrevHandleDNDPt);
	}
    }
}

/******************************************************************************
 WaitForLastStatusMsg (private)

	Returns True if it receives XdndStatus message with accept==True.

 ******************************************************************************/

Bool
JXDNDManager: :WaitForLastStatusMsg ()
{
    if (!itsReceivedStatusFlag) {
	return False;
    } else if (!itsWaitForStatusFlag) {
	return itsWillAcceptDropFlag;
    }
    itsSendHereMsgFlag = False;	/* don't send any more messages */

    JXSelectionManager *selMgr = itsDisplay->GetSelectionManager ();

    Atom messageList[] =
    {1, itsDNDStatusXAtom};

    XEvent xEvent;
    int endTime = clock () + kWaitForLastStatusTime;
    while (clock () < endTime) {
	if (XCheckIfEvent (*itsDisplay, &xEvent, GetNextStatusEvent,
			   static_cast (char *, messageList))) {
	    if (xEvent.type == ClientMessage) {
		const Bool ok = HandleClientMessage (xEvent.xclient);
		assert (ok);
		if (xEvent.xclient.message_type == itsDNDStatusXAtom) {
		    return itsWillAcceptDropFlag;
		}
	    } else if (xEvent.type == SelectionRequest) {
		selMgr->HandleSelectionRequest (xEvent.xselectionrequest);
	    }
	    /* reset timeout */

	    endTime = clock () + kWaitForLastStatusTime;
	}
    }

    return False;
}

/* static */

Bool
JXDNDManager: :GetNextStatusEvent
(
    Display * display,
    XEvent * event,
    char *arg
) {
    Atom *messageList = static_cast (Atom *, arg);
    const int atomCount = messageList[0];

    if (event->type == ClientMessage) {
	for (JIndex i = 1; i <= atomCount; i++) {
	    if ((event->xclient).message_type == messageList[i]) {
		return True;
	    }
	}
    } else if (event->type == SelectionRequest) {
	return True;
    }
    return False;
}

/******************************************************************************
 ReceiveWithFeedback (virtual protected)

	This catches:
		1)  target crashes
		2)  source crashes while a message is in the network

 ******************************************************************************/

void
 JXDNDManager::ReceiveWithFeedback
 (
     JBroadcaster * sender,
     Message * message
) {
  if (sender == itsDisplay && message->Is (JXDisplay: :kXError)) {
      JXDisplay: : XError * err = dynamic_cast (JXDisplay: :XError *, message);
	assert (err != NULL);

	/* source: target crashed -- nothing to do */

	/* target: source crashed */

	if (itsDragger == NULL && itsDraggerWindow != None &&
	    err->GetType () == BadWindow &&
	    err->GetXID () == itsDraggerWindow) {
#if JXDND_DEBUG_MSGS
	    cout << "JXDND: Source crashed via XError" << endl;
#endif

	    HandleDNDLeave1 ();
	    err->SetCaught ();
	}
	/* target: residual from source crash -- nothing to do */
    }
}

/******************************************************************************
 HandleDestroyNotify 

	This catches source crashes while it is processing XdndStatus.

 ******************************************************************************/

Bool
JXDNDManager: :HandleDestroyNotify
(
    const XDestroyWindowEvent & xEvent
) {
    if (itsDragger == NULL && itsDraggerWindow != None &&
	xEvent.window == itsDraggerWindow) {
#if JXDND_DEBUG_MSGS
	cout << "JXDND: Source crashed via DestroyNotify" << endl;
#endif

	HandleDNDLeave1 ();
	return True;
    } else {
	return False;
    }
}
